/*
 * Copyright (c) 2010 The Trustees of University of Illinois. 
 *                    All rights reserved.
 *
 * Author(s): Torsten Hoefler <htor@illinois.edu>
 *
 */

#include <iostream>
#include <string.h>
#include <vector>
#include <algorithm>
#include <numeric>
#include "MersenneTwister.h"
#include <fftw3.h>
#include <mpi.h>
#include "fft2d_cmdline.h"


using namespace std;

typedef struct { double r,i; } base_t;

static int n=800;
static int check=1;
static bool dbg_print=0;

/*
 * layout for 3x3: (0,0) (0,1) (0,2) (1,0) (1,1) (1,2) (2,0) (2,1) (2,2)
 */
static inline int idx(int x, int y) {
  return n*x+y;
}


void print(base_t *arr, int size, std::string str) {
  if(dbg_print) { 
    int p, r;
    MPI_Comm_size(MPI_COMM_WORLD, &p);
    MPI_Comm_rank(MPI_COMM_WORLD, &r);

    base_t* narr = (base_t*)malloc(sizeof(base_t)*size*p);
    MPI_Gather(&arr[0], 2*size, MPI_DOUBLE, &narr[0], 2*size, MPI_DOUBLE, 0, MPI_COMM_WORLD);

    if(!r) {
      cout << "----------------- " << str << endl ;
      for(int pp=0; pp<p; ++pp) {
        cout << "[" << pp << "] ";
        for(int i=0; i<size; ++i) {
          cout << "(" << narr[pp*size+i].r << "," << narr[pp*size+i].i << ") ";
        }
        cout << endl;
      }
    }
    free(narr);
  }
}

int main(int argc, char **argv ) {


  MPI_Init(NULL,NULL);
  MTRand mt;

  int p,r;
  MPI_Comm comm=MPI_COMM_WORLD;
  MPI_Comm_size(comm, &p);
  MPI_Comm_rank(comm, &r);
  //parse cmdline arguments
  struct fft2d_cmd_struct args_info;
  if (fft2d_parser(argc, argv, &args_info) != 0) {
    std::cout << "parser error\n";
  }

  int flops=1;
  double flopsx;
  dbg_print = args_info.debug_arg;
  n = args_info.size_arg;
  check = args_info.check_arg;
  flops = args_info.flops_arg;

 if(!r) cout << "p=" << p << ", n=" << n << endl;

  if(n%p != 0) { 
    cout << "n (" << n << ") must be a multiple of p (" << p << ")\n";
    MPI_Abort(comm, 1);
  }


  for(int iter=0; iter<= 1; iter++) {
    if(!r) cout << " ----- iteration " << iter << " -------\n";
    for(int send_datatype = 0; send_datatype <= 1; send_datatype++) {
      for(int recv_datatype = 0; recv_datatype <= 1; recv_datatype++) {

    flopsx=0.0;
    
    MPI_Datatype complex_t, dtype_gather_t, dtype_scatter_t;
    MPI_Type_contiguous(2, MPI_DOUBLE, &complex_t);
    MPI_Type_commit(&complex_t);

    
    std::vector<base_t> in(n*n/p), out1(n*n/p), out2(n*n/p);
    if(check) {
      in.resize(n*n);
      out1.resize(n*n);
      out2.resize(n*n);
    }

    int up;
    if(check) up = n;
    else up=n/p;

    for(int x=0; x<up; ++x) {
      for(int y=0; y<n; ++y) {
        if(!dbg_print) {
          in[idx(x,y)].r = mt.rand();
          in[idx(x,y)].i = mt.rand();
        } else {
          in[idx(x,y)].r = x;
          in[idx(x,y)].i = y;
        }
    } }

    if(check) {
      if(!r) MPI_Scatter(&in[0], n*n/p, complex_t, MPI_IN_PLACE, n*n/p, complex_t, 0, comm);
      else MPI_Scatter(&in[0], n*n/p, complex_t, &in[0], n*n/p, complex_t, 0, comm);
    }

    std::copy(in.begin(), in.end(), out2.begin());

    fftw_plan plan = fftw_plan_dft_1d(n, (double(*)[2])&out2[0], (double(*)[2])&out2[0], FFTW_FORWARD, FFTW_ESTIMATE); // same buffer
    if(flops) {
      double add, mul, fma;
      fftw_flops(plan, &add, &mul, &fma);
      flopsx += add + mul + 2*fma;
    }

    if(recv_datatype) { 
      MPI_Datatype dtype_vector_t, dtype_resized_t;
      MPI_Type_vector(n/p /* count */, 1 /* blocklength */, n /* stride */, complex_t, &dtype_vector_t);
      // this seems a bit hack-ish to trick the datatype engine into overlapping datatypes 
      MPI_Type_create_resized(dtype_vector_t, 0 /* lb */, sizeof(base_t) /* extent */, &dtype_resized_t);
      MPI_Type_contiguous(n/p, dtype_resized_t, &dtype_scatter_t);
      MPI_Type_commit(&dtype_scatter_t);
    }
    if(send_datatype) { 
      MPI_Datatype dtype_vector_t;
      MPI_Type_vector(n/p /* count */, n/p /* blocklength */, n /* stride */, complex_t, &dtype_vector_t);
      // this seems a bit hack-ish to trick the datatype engine into overlapping datatypes 
      MPI_Type_create_resized(dtype_vector_t, 0 /* lb */, n/p*sizeof(base_t) /* extent */, &dtype_gather_t);
      MPI_Type_commit(&dtype_gather_t);
    }


    ///////////////////////////////////////////////////////////////////////////////////
    // timing starts here! 
    double t=-MPI_Wtime();
    
    // 1-d FFTs
    // layout for 3x3: (0,0) (0,1) (0,2) (1,0) (1,1) (1,2) (2,0) (2,1) (2,2)
    for(int x=0; x<n/p; ++x) {
      if(!dbg_print) {
        fftw_execute_dft(plan, (double(*)[2])&out2[idx(x,0)], (double(*)[2])&out2[idx(x,0)]);
        /*plan = fftw_plan_dft_1d(n, (double(*)[2])&out2[idx(x,0)], (double(*)[2])&out2[idx(x,0)], FFTW_FORWARD, FFTW_ESTIMATE);
        fftw_execute(plan);
        fftw_destroy_plan(plan);*/
      } 
    }

    print(&out2[0], n*n/p, "original");

    if(!send_datatype) {
      // n/p blocks of size n/p with stride n
      for(int pp=0; pp<p; ++pp) {
        for(int x=0; x<n/p; ++x) {
          memcpy(&out1[pp*n/p*n/p + x*n/p], &out2[pp*n/p+ x*n], n/p*sizeof(base_t));
          //for(int y=0; y<n/p; ++y) {
          //  out1[pp*n/p*n/p + x*n/p + y] = out2[pp*n/p+ x*n + y];
      } } //}
      
      print(&out1[0], n*n/p, "before alltoall");
    } 

    /* TODO: send and recv datatype work separatly but not together :-) (MPI_IN_PLACE would not work here) -- fix array problems */
    if(send_datatype && recv_datatype) MPI_Alltoall(&out2[0], 1, dtype_gather_t, &out1[0], 1, dtype_scatter_t, comm);
    if(!send_datatype && recv_datatype) MPI_Alltoall(&out1[0], n/p*n/p, complex_t, &out2[0], 1, dtype_scatter_t, comm);
    if(!send_datatype && !recv_datatype) MPI_Alltoall(&out1[0], n/p*n/p, complex_t, &out2[0], n/p*n/p, complex_t, comm);
    if(send_datatype && !recv_datatype) MPI_Alltoall(&out2[0], 1, dtype_gather_t, &out1[0], n/p*n/p, complex_t, comm);
    
    if(!recv_datatype) {
      print(&out1[0], n*n/p, "after alltoall");

      // local transpose to layout for 3x3: (0,0) (1,0) (2,0) (0,1) (1,1) (2,1) (0,2) (1,2) (2,2)
      for(int x=0; x<n/p; ++x) {
        for(int y=0; y<n; ++y) {
          if(!send_datatype && !recv_datatype) out1[x*n + y] = out2[x + y*n/p];
          else                                  out2[x*n + y] = out1[x + y*n/p];
          //if(!r) cout << x*n + y << endl;
      } }
    } 

    print(&out2[0], n*n/p, "after unpack");
    
    // fft in y-direction (which is transposed)
    for(int x=0; x<n/p; ++x) {
      if(!dbg_print) {
        if (send_datatype && recv_datatype ||
            !send_datatype && !recv_datatype) fftw_execute_dft(plan, (double(*)[2])&out1[idx(x,0)], (double(*)[2])&out1[idx(x,0)]);
        else                                  fftw_execute_dft(plan, (double(*)[2])&out2[idx(x,0)], (double(*)[2])&out2[idx(x,0)]);
        /*if (send_datatype && recv_datatype ||
            !send_datatype && !recv_datatype) plan = fftw_plan_dft_1d(n, (double(*)[2])&out1[idx(x,0)], (double(*)[2])&out1[idx(x,0)], FFTW_FORWARD, FFTW_ESTIMATE);
        else                                  plan = fftw_plan_dft_1d(n, (double(*)[2])&out2[idx(x,0)], (double(*)[2])&out2[idx(x,0)], FFTW_FORWARD, FFTW_ESTIMATE);
        fftw_execute(plan);
        fftw_destroy_plan(plan);*/
      }
    }

    if(!send_datatype) {
      for(int pp=0; pp<p; ++pp) {
        for(int x=0; x<n/p; ++x) {
          //for(int y=0; y<n/p; ++y) {
          //  if (!send_datatype && !recv_datatype) out2[pp*n/p*n/p + x*n/p + y] = out1[pp*n/p + x*n + y];
          //  else                                  out1[pp*n/p*n/p + x*n/p + y] = out2[pp*n/p + x*n + y];
          //  //if(!r) cout << pp*n/p*n/p + x*n/p + y << endl;
          //} 
          if (!send_datatype && !recv_datatype) memcpy(&out2[pp*n/p*n/p + x*n/p], &out1[pp*n/p + x*n] ,n/p*sizeof(base_t));
          else                                  memcpy(&out1[pp*n/p*n/p + x*n/p], &out2[pp*n/p + x*n] ,n/p*sizeof(base_t));
      } }

      print(&out1[0], n*n/p, "before alltoall");
      
      // transpose back to layout for 3x3: (0,0) (1,0) (2,0) (0,1) (1,1) (2,1) (0,2) (1,2) (2,2)
      //MPI_Alltoall(MPI_IN_PLACE, n/p*n/p, complex_t, &out1[0], n/p*n/p, complex_t, comm);
    } 

    if(send_datatype && recv_datatype) MPI_Alltoall(&out1[0], 1, dtype_gather_t, &out2[0], 1, dtype_scatter_t, comm);
    if(!send_datatype && recv_datatype) MPI_Alltoall(&out1[0], n/p*n/p, complex_t, &out2[0], 1, dtype_scatter_t, comm);
    if(!send_datatype && !recv_datatype) MPI_Alltoall(&out2[0], n/p*n/p, complex_t, &out1[0], n/p*n/p, complex_t, comm);
    if(send_datatype && !recv_datatype) MPI_Alltoall(&out2[0], 1, dtype_gather_t, &out1[0], n/p*n/p, complex_t, comm);

    if(!recv_datatype) {
      print(&out1[0], n*n/p, "after alltoall");

      // n blocks of size n/p with stride n/p
      for(int x=0; x<n/p; ++x) {
        for(int y=0; y<n; ++y) {
          out2[x*n + y] = out1[x + y*n/p];
          //if(!r) cout << x*n + y << endl;
      } }
    }

    t += MPI_Wtime();
    // timing ends here! 
    ///////////////////////////////////////////////////////////////////////////////////

    fftw_destroy_plan(plan);

    print(&out2[0], n*n/p, "after unpack");

    if(check) {
      MPI_Gather(&out2[0], n*n/p, complex_t, &out1[0], n*n/p, complex_t, 0, comm);

      // 2-d FFT
      plan = fftw_plan_dft_2d(n, n, (double(*)[2])&in[idx(0,0)], (double(*)[2])&out2[idx(0,0)], FFTW_FORWARD, FFTW_ESTIMATE);
      fftw_execute(plan);
      fftw_destroy_plan(plan);

      base_t sum1 = {0.0,0.0}; 
      base_t sum2 = {0.0,0.0}; 
      for(int x=0; x<n; ++x) {
        for(int y=0; y<n; ++y) {
          sum1.r += out1[idx(x,y)].r;
          sum1.i += out1[idx(x,y)].i;
          sum2.r += out2[idx(x,y)].r;
          sum2.i += out2[idx(x,y)].i;
      } }

      if(!r && !dbg_print) std::cout << "diff: "<< sum1.r - sum2.r << ", " << sum1.i - sum2.i << "\n";
    }

    if(!r) cout << "time: " <<  t << "; stype, rtype: " << send_datatype<< ", " << recv_datatype<< endl;
    if(!r && flops) cout << "flops: x:" <<  flopsx << "; y: " << flopsx << "; GFlop/s total: " << (n/p*flopsx+n/p*flopsx)/t/10e9 << " (overall: " << (n/p*flopsx+n/p*flopsx)/t/10e9*p << ")" << endl; 

  } } }

  MPI_Finalize();

  return 0;
}
